'''For generating human-readable string representations of other Python objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any

from textwrap import indent
from enum import StrEnum


class Justification(StrEnum):
    '''For specifying string justification'''
    LEFT   = '<'
    CENTER = '^'
    RIGHT  = '>'
Just = Justification # alias for the lazy or hurried

def procrustean_string(
        string : str,
        length : int,
        padding : str=' ',
        just : Justification=Justification.LEFT,
    ) -> int:
    '''Takes a string and a target length and returns a new string which begins
    with the same characters as the original string but is clamped to the target length,
    truncating or padding if the original string is too long or short, respectively
    
    Parameters
    ----------
    string : str
        The string to stretch or cut
    length : int
        The target number of characters in the final string
    padding : str, default=" "
        A single character which shold be used as padding 
        when strings are too short, by default just a space
        MUST BE EXACTLY ONE CHARACTER!
    just : Justification, default=Justification.LEFT
        Enum specifier of how to justify a padded string
        Options are Justification.LEFT, Justification.CENTER, or Justification.RIGHT  
        
    Returns
    -------
    fmt_str : str
        A string which begins with the same characters as "string" but has
        precisely the specified length, with specified padding as specified
    '''
    if not (isinstance(length, int) and (length >= 0)):
        raise ValueError(f'Target string length must be a non-negative integer (not {length})')
    if not len(padding) == 1:
        raise IndexError(f'Padding string must contain exactly one character (passed "{padding}")')
        
    return f'{string[:length]:{padding}{just.value}{length}}'

def dict_to_indented_str(dict_to_stringify : dict[Any, Any], level_delimiter : str='\t', line_sep : str='\n') -> str:
    '''Generate a pretty-printable string from a (possibly nested) dictionary,
    with each level of nesting indicated by "level_delimiter"'''
    text = []
    for key, value in dict_to_stringify.items():
        if isinstance(value, dict):
            text.append(f'{key!r}')
            text.append(indent(dict_to_indented_str(value), level_delimiter)) # recursive call for nested dicts
        else:
            text.append(f'{key!r} : {value!r}') # call repr methods

    return line_sep.join(text)


'''For querying chemical databases for information about molecules specified by SMILES string and other structures'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER  = logging.getLogger(__name__)

from typing import Any, ClassVar, Container, Iterable, Optional, Sequence
from abc import ABC, abstractmethod

import requests

from ..genutils.decorators.classmod import register_abstract_class_attrs, register_subclasses
from ..genutils.importutils.dependencies import requires_modules, MissingPrerequisitePackage


# CUSTOM EXCEPTIONS
class InvalidPropertyError(Exception):
    '''Raised when attempting to query a property that a chemical database service cannot provide'''
    pass

class NullPropertyResponse(Exception):
    '''Raised when a chemical database query doesn't fail BUT returns a NoneType where not allowed'''
    pass

class ChemicalDataQueryFailed(Exception):
    '''Raised when a chemical data query is unfulfilled by a service'''
    pass

# STRATEGIES BASE FOR QUERYING CHEMICAL DATA
@register_subclasses(key_attr='service_name')
@register_abstract_class_attrs('service_name')
class ChemDBServiceQueryStrategy(ABC):
    '''Implementation of queries from a particular chemical database'''
    @abstractmethod
    def _get_property(self, prop_name : str, representation : str, **kwargs) -> Optional[Any]:
        ...
        
    @classmethod
    def dependencies(cls) -> Iterable[str]:
        '''For internals, allows dynamic checking for package dependencies (useful for automating unit test boilerplate)'''
        ...
        
    @classmethod
    @abstractmethod
    def is_online(cls) -> bool:
        '''Check if the service being queried is online and can accept requests'''
        ...
        
    @classmethod
    @abstractmethod
    def queryable_properties(cls) -> Container[str]:
        '''List which chemical property names can be queried from the service'''
        ...

    @classmethod
    @abstractmethod
    def queryable_namespaces(cls) -> Container[str]:
        '''List which chemical identification types can be searched through by the service'''
        ...
        
    def validate_property(self, prop_name : str) -> None:
        '''Pre-check to ensure that a property is queryable from a service before attempting HTTP query'''
        if prop_name not in self.queryable_properties():
            prop_options_str = '\n'.join(sorted(self.queryable_properties()))
            prop_error_msg = f'Cannot query property "{prop_name}" from {self.service_name}'
            LOGGER.error(prop_error_msg) # log briefer error message in cases where the ensuing ValueError is bypassed
            
            raise InvalidPropertyError(f'{prop_error_msg};\nChoose from one of the following property names:\n{prop_options_str}')
        
    def get_property(
            self, 
            prop_name : str, 
            representation : str, 
            namespace : Optional[str],
            keep_first_only : bool=True,
            allow_null_return : bool=False,
            **kwargs
        ) -> Optional[Any]:
        '''Fetch a property associated with a molecule from a chemical database query service'''
        LOGGER.info(f'Sent query request for property "{prop_name}" to {self.service_name}')
        self.validate_property(prop_name=prop_name)
        
        prop_val = self._get_property(prop_name=prop_name, representation=representation, namespace=namespace, **kwargs)
        if not prop_val:
            prop_val = None # cast empty lists, strings, etc to NoneType
        
        if isinstance(prop_val, Container) and not isinstance(prop_val, str) and keep_first_only: # avoid bug where first char of string response is returned
            prop_val = prop_val[0]
        
        if (prop_val is None) and (not allow_null_return): # NOTE: duplicated NoneType check is needed to catch empty containers which are cast to None above
            null_error_msg = f'{self.service_name} returned NoneType "{prop_name}", which is declared invalid by call signature'
            LOGGER.error(null_error_msg)
            
            raise NullPropertyResponse(null_error_msg)
        LOGGER.info(f'Successfully received property "{prop_name}" from {self.service_name}')
                
        return prop_val
    
# CONCRETE IMPLEMENTATIONS OF CHEMICAL DATABASE SERVICE QUERIES
## NIH CACTUS
cirpy_error = MissingPrerequisitePackage(
    importing_package_name=__spec__.name,
    use_case='Querying the NIH CACTUS Chemical Identifier Resolver (CIR)',
    install_link='https://cirpy.readthedocs.io/en/latest/guide/install.html',
    dependency_name='cirpy',
    dependency_name_formal='CIRpy',
)

class NIHCACTUSQueryStrategy(ChemDBServiceQueryStrategy):
    '''
    Implementation of chemical query requests to the NIH's CADD group 
    Cheminformatics Tools and User Services (CACTUS) Chemical Identifier Resolver (CIR)
    '''
    service_name : ClassVar[str] = 'NIH CACTUS CIR'
    
    @classmethod
    def dependencies(cls):
        return ['cirpy']
    
    @classmethod
    @requires_modules('cirpy', missing_module_error=cirpy_error)
    def queryable_properties(cls) -> set[str]:
        import cirpy 
        
        _CIR_PROPS = {  # see official docs for more info: https://cactus.nci.nih.gov/chemical/structure_documentation
            'smiles',
            'ficts',
            'ficus',
            'uuuuu',
            'hashisy',
            'names',
            'iupac_name',
            'cas',
            'chemspider_id',
            'image',
            'twirl',
            'mw',
            'formula',
            'h_bond_donor_count',
            'h_bond_acceptor_count',
            'h_bond_center_count',
            'rule_of_5_violation_count',
            'rotor_count',
            'effective_rotor_count',
            'ring_count',
            'ringsys_count',
            'inchi',
            'inchikey',
            # shortened aliases of InChI-related properties
            'stdinchi', 
            'stdinchikey',
            # these were not documented on CACTUS or by cirpy, but scraped from webchem: https://github.com/ropensci/webchem/blob/master/R/cir.R#L168-L174
            'deprotonable_group_count',
            'heavy_atom_count',
            'heteroatom_count',
            'hydrogen_atom_count',
            'monoisotopic_mass',
            'protonable_group_count',
            'xlogp2',
        }
        return _CIR_PROPS | cirpy.FILE_FORMATS
    
    @classmethod
    def is_online(cls):
        response = requests.head('https://cactus.nci.nih.gov/chemical/structure')
        return response.status_code < 500 # NOTE: could also be more stringent and check == 200 for OK; enough to just check server-side error for now
    
    @classmethod
    def queryable_namespaces(cls) -> set[str]:
        return { # obtained from https://cirpy.readthedocs.io/en/latest/guide/resolvers.html
            'smiles',
            'stdinchikey',
            'stdinchi',
            'ncicadd_identifier', # (for FICTS, FICuS, uuuuu)
            'hashisy',
            'cas_number',
            'name', # this is not documented but DOES work
            'name_by_opsin',
            'name_by_cir',
        }
    
    @requires_modules('cirpy', missing_module_error=cirpy_error)
    def _get_property(self, prop_name : str, representation : str, namespace : Optional[str]=None, **kwargs):
        import cirpy
        
        return cirpy.resolve(representation, prop_name, resolvers=[namespace], **kwargs)

## PubChem
pubchempy_error = MissingPrerequisitePackage(
    importing_package_name=__spec__.name,
    use_case='Querying the PubChem Compound database',
    install_link='https://pubchempy.readthedocs.io/en/latest/guide/install.html',
    dependency_name='pubchempy',
    dependency_name_formal='PubChemPy',
)

class PubChemQueryStrategy(ChemDBServiceQueryStrategy):
    '''
    Implementation of chemical query requests to PubChem via the
    PUG REST API (https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest)
    '''
    service_name : ClassVar[str] = 'PubChem'
    
    @classmethod
    def dependencies(cls):
        return ['pubchempy']
    
    @classmethod
    def is_online(cls):
        response = requests.get('https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/aspirin/property/IUPACName/TXT') # sample query which is well-formatted
        return response.status_code < 500 # NOTE: enough to just check server-side error for now, but could be more stringent and check if ==200
        
    @classmethod
    @requires_modules('pubchempy', missing_module_error=pubchempy_error)
    def queryable_properties(cls) -> set[str]:
        from pubchempy import PROPERTY_MAP
        
        return set(PROPERTY_MAP.keys()) | set(PROPERTY_MAP.values()) | {'Fingerprint2D'} # also taken from webchem: https://github.com/ropensci/webchem/blob/master/R/pubchem.R#L377C21-L392C55
    
    @classmethod
    def queryable_namespaces(cls) -> set[str]:
        return { # obtained from https://pubchem.ncbi.nlm.nih.gov/docs/pug-rest#section=Input
            'cid',
            'name',
            'smiles',
            'inchi',
            'sdf',
            'inchikey',
            'formula',
            'listkey',
        }
    
    @requires_modules('pubchempy', missing_module_error=pubchempy_error)
    def _get_property(self, prop_name : str, representation : str, namespace : Optional[str]='smiles', **kwargs) -> Optional[Any]:
        from pubchempy import PROPERTY_MAP, get_properties, PubChemPyError
        
        official_prop_name = PROPERTY_MAP.get(prop_name, prop_name) # this is done internally, but needed here to extract the property value from the final return dict
        try:
            pubchem_result = get_properties(official_prop_name, identifier=representation, namespace=namespace, **kwargs)
        except PubChemPyError:
            raise requests.HTTPError # discards some information in return for making Strategy interface oblivious to pubchempy (i.e. in case it is not installed)
        else:
            if pubchem_result:
                pubchem_result = [
                    query_result[official_prop_name] # extract property value from extraneous CID (and any other) info
                        for query_result in pubchem_result
                            if official_prop_name in query_result # skip if return doesn't contain the info we specifically requested (happens occasionally for some reason)
                ] 
            return pubchem_result
        
# UTILITY FUNCTIONS EMPLOYING GENERIC STRATEG(Y/IES)
def get_chemical_property(
        prop_name : str, 
        representation : str, 
        namespace : str='smiles',
        keep_first_only : bool=True,
        allow_null_return : bool=False,
        fail_quietly : bool=False,
        services : Optional[Sequence['ChemDBServiceQueryStrategy']]=None,
        **kwargs,
    ) -> Optional[Any]:
    '''Attempt to fetch a molecular property from a variety of chemical database services, either
    provided manually (in the order they should be checked) or ALL implemented service queries by default
    
    Will return the first valid returned result or, if all services fail, raise Exception
    '''
    # determine services which should be queried
    if services is None:
        services = [chem_query_strat_type() for chem_query_strat_type in ChemDBServiceQueryStrategy.subclass_registry.values()]
    if not services: # check if "services" turns out to be an empty collection (either as-passed or because no subclasses are implemented when defaulting)
        raise IndexError('Must provide at least one chemical database querying strategy to "services"')
    n_services_to_try : int = len(services)
    
    # query services sequentially in order of appearance
    for i, service in enumerate(services, start=1):
        ## validate type of service strategies
        if isinstance(service, type):
            service = service() # allows ChemDBServiceQueryStrategy types to be passed in lieu of instances
        if not isinstance(service, ChemDBServiceQueryStrategy):
            raise TypeError(f'Services must be specified as {ChemDBServiceQueryStrategy.__name__} instances, not objects of type {type(service.__name)}')
        
        ## attempt to query result from service
        LOGGER.info(f'Attempting chemical property query to service {i}/{n_services_to_try} ("{service.service_name}"):')
        try:
            prop_val = service.get_property(
                prop_name,
                representation,
                namespace,
                keep_first_only=keep_first_only,
                allow_null_return=allow_null_return,
                **kwargs,
            )
            return prop_val
        except requests.HTTPError:
            LOGGER.error(f'Query to {service.service_name} failed, either due to connection timeout or invalid request')
            continue
        except (InvalidPropertyError, NullPropertyResponse): # skip over invalid property names (keep trying other services rather than failing)
            # log messages baken in to respective raises for these custom exceptions
            continue
    else: # take action when None of the provided services turn up fruitful
        fail_msg = 'Query could not be fulfilled by any of the provided chemical query services'
        if fail_quietly:
            LOGGER.error(f'{fail_msg}; returning NoneType')
            return None
        else: # fail vocally if none of the services can fulfill the property request
            raise ChemicalDataQueryFailed(fail_msg)


'''Unit tests for `dependencies` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

from typing import Any, Callable, Union
from polymerist.genutils.importutils import dependencies


# Testing module finding
@pytest.mark.parametrize(
    'module_names, expected_found', [
        (['polymerist'], True),     # we'd better hope the parent module is present if we're running tests on it :P
        (['sys'], True),            # test stdlib packages which ought to be present if Python is
        (['os', 'sys'], True),      # test that unpacking also works
        (['fake--module'], False),  # test an obviously fake module name (don't want to try an actual module in case it becomes an dependency someday)
        ([42], False),              # test something that isn't even a module to check error handling
    ]
)
def test_modules_installed(module_names : list[str], expected_found : bool) -> None:
    '''Check that module install checker correctly identifies present and absent modules'''
    assert dependencies.modules_installed(*module_names) == expected_found
    
# Testing requires_modules() decorator
@pytest.mark.parametrize(
    'module_name,missing_module_error',
    [
        ('os', ImportError),
        ('os', ImportError('This is not the default message!')),
        pytest.param(
            'os', 42, # note that module IS valid here but the Exception raised is not (meta-error)
            marks=pytest.mark.xfail(
                raises=TypeError, 
                reason='Non Exception-like object passed to missing_module_error',
                strict=True,
            )
        ),
        pytest.param(
            'fake--module', ImportError, 
            marks=pytest.mark.xfail(
                raises=ImportError, # NOTE that this ImportError MUST be the same as the import error passed via args
                reason='The required module shouldn\'t be found in the environment',
                strict=True,
            )
        ),
        pytest.param(
            'fake--module', AttributeError('something else'), 
            marks=pytest.mark.xfail(
                raises=AttributeError, # NOTE that this AttributeError MUST be the same as the import error passed via args
                reason='The required module shouldn\'t be found in the environment',
                strict=True,
            )
        ),
    ]
)
def test_requires_modules(module_name : str, missing_module_error : Union[Exception, type[Exception]]) -> None:
    '''Test that the requires_modules decorator correctly wraps functions'''
    @dependencies.requires_modules(module_name, missing_module_error=missing_module_error)
    def func() -> str:
        return 'I am pointless'
    
    _ = func() # no assertion needed, xfail cases should raise Exception while working cases will ternimate without Exception


'''Unit tests for `chemdbqueries` package'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest

from typing import Any
from dataclasses import dataclass, asdict

from requests import HTTPError

from polymerist.genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage
from polymerist.smileslib.chemdbqueries import (
    get_chemical_property,
    InvalidPropertyError,
    NullPropertyResponse,
    ChemicalDataQueryFailed,
    ChemDBServiceQueryStrategy,
    # NOTE: these strageies are implemented to be defined even if the packages in question aren't installed
    # it's just that instances will raise excception on most of their method calls
    NIHCACTUSQueryStrategy,
    PubChemQueryStrategy,
    get_chemical_property,
    
)

CHEMDB_STRATEGY_ONLINE : dict[str, bool] = {}
CHEMDB_STRATEGY_DEPENDENCIES_MET : dict[ChemDBServiceQueryStrategy, bool] = {}
for ChemDBStrategy in ChemDBServiceQueryStrategy.__subclasses__(): # dynamically determine criteria for which services should be tested
    CHEMDB_STRATEGY_ONLINE[          ChemDBStrategy] = ChemDBStrategy.is_online()
    CHEMDB_STRATEGY_DEPENDENCIES_MET[ChemDBStrategy] = modules_installed(*ChemDBStrategy.dependencies())

@pytest.mark.parametrize(
    'service_type', 
    [
        pytest.param(
            service_type,
            marks=pytest.mark.xfail(
                raises=MissingPrerequisitePackage,
                reason='Unsatisfied dependency needed for chemical database service to be imported',
                strict=True,
            )
        )
            for service_type, dependencies_met in CHEMDB_STRATEGY_DEPENDENCIES_MET.items()
                if not dependencies_met
    ]
)
def test_missing_dependency_xfail(service_type : ChemDBServiceQueryStrategy) -> None:
    '''Test whether checks for missing prerequisite dependencies are active'''
    _ = service_type.queryable_properties() # happen to know this requires respective dependencies at the time of writing


@dataclass
class ChemDBQueryExample:
    '''For encapsulating the many parameters passable to a chemical database service query'''
    prop_name : str
    representation : str
    namespace : str
    keep_first_only : bool
    allow_null_return : bool

# SERVICE_QUERY_EXAMPLES :

EXAMPLES_BY_SERVICE : list[tuple[ChemDBServiceQueryStrategy, ChemDBQueryExample, Any]] = [
    # for NIH CACTUS
    ( ## simple queries known to work for all services
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample( 
            'iupac_name',
            'CCO',
            namespace='smiles',
            keep_first_only=True,
            allow_null_return=False,
        ),
        'ethanol'
    ),
    (
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample( 
            'inchi',
            'N-methylformamide',
            namespace='name',
            keep_first_only=True,
            allow_null_return=False,
        ),
        'InChI=1/C2H5NO/c1-3-2-4/h2H,1H3,(H,3,4)/f/h3H',
    ),
    ( ## testing that different namespaces can be queried
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample(
            'mw',
            'benzophenone',
            namespace='name', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        '182.2214'
    ),
    ( ## testing that returns with multiple data values work
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample(
            'names',
            'c1ccccc1-C(=S)S',
            namespace='smiles', 
            keep_first_only=False, 
            allow_null_return=False,
        ),
        ['Benzenecarbodithioic acid', '121-68-6', 'UPCMLD00WV-104', 'EINECS 204-491-4', 'Benzenecarbodithioic acid', 'Dithiobenzoic acid', 'NSC732246']
    ),
    ( ## testing that enabling and disabling None returns is handled properly in both cases
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample(
            'inchi',
            'bogus-name', # this is obviously fake and should not return anything
            namespace='name', 
            keep_first_only=True,
            allow_null_return=True,
        ),
        None
    ),
    pytest.param( 
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample(
            'inchi',
            'bogus-name', # this is obviously fake and should not return anything
            namespace='name', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        None,
        marks=pytest.mark.xfail(
            raises=(NullPropertyResponse, ChemicalDataQueryFailed),
            reason='Did not allow response to be NoneType',
            strict=True,
        )
    ),
    pytest.param( ## testing that invalid property values are caught before attempting a query
        NIHCACTUSQueryStrategy,
        ChemDBQueryExample(
            'in_no_way_a_valid_property', # this should not even be considered a valid property
            'benophenone', 
            namespace='name', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        None,
        marks=pytest.mark.xfail(
            raises=(InvalidPropertyError, ChemicalDataQueryFailed),
            reason='Tried to query a property that does not exist',
            strict=True,
        )
    ),
    
    # for PubChem
    ( ## simple queries known to work for all services
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'iupac_name',
            'CCO',
            namespace='smiles',
            keep_first_only=True,
            allow_null_return=False
        ),
        'ethanol'
    ),
    (
        PubChemQueryStrategy,
        ChemDBQueryExample( 
            'inchi',
            'N-methylformamide',
            namespace='name',
            keep_first_only=True,
            allow_null_return=False,
        ),
        'InChI=1S/C2H5NO/c1-3-2-4/h2H,1H3,(H,3,4)',
    ),
    ( ## testing that different namespaces can be queried
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'MolecularWeight',
            'InChI=1S/C2H5NO/c1-3-2-4/h2H,1H3,(H,3,4)',
            namespace='inchi', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        '59.07',
    ),
    ( ## testing that returns with multiple data values work
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'HeavyAtomCount',
            'CCO', 
            namespace='smiles', 
            keep_first_only=False,
            allow_null_return=False,
        ),
        [3], # note that this is wrapped in a list, as are all PubChem queries by deualt; I couldn't find a good example which returns more than one value like cirpy does
    ),
    pytest.param( ## testing sending malformed queries to PubChem
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'inchi',
            'bogus-name', # this is obviously fake and should not return anything
            namespace='smiles', 
            keep_first_only=True,
            allow_null_return=True,
        ),
        None,
        marks=pytest.mark.xfail(
            raises=(HTTPError, ChemicalDataQueryFailed),
            reason='Invalid request sent to PubChem (queried a name as a SMILES string)',
            strict=True,
        )
    ),
    ( ## testing that enabling and disabling None returns is handled properly in both cases
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'inchi',
            'bogus-name', # this is obviously fake and should not return anything
            namespace='name', 
            keep_first_only=True,
            allow_null_return=True,
        ),
        None,
    ),
    pytest.param(  
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'inchi',
            'bogus-name', # this is obviously fake and should not return anything
            namespace='name', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        None,
        marks=pytest.mark.xfail(
            raises=(NullPropertyResponse, ChemicalDataQueryFailed),
            reason='Did not allow response to be NoneType',
            strict=True,
        )
    ),
    pytest.param( ## testing that invalid property values are caught before attempting a query
        PubChemQueryStrategy,
        ChemDBQueryExample(
            'in_no_way_a_valid_property', # this should not even be considered a valid property
            'benophenone', 
            namespace='name', 
            keep_first_only=True,
            allow_null_return=False,
        ),
        None,
        marks=pytest.mark.xfail(
            raises=(InvalidPropertyError, ChemicalDataQueryFailed),
            reason='Tried to query a property that does not exist',
            strict=True,
        )
    ),
]
    
@pytest.mark.parametrize('service_type,query_example,expected_return', EXAMPLES_BY_SERVICE)
class TestChemicalDatabaseServiceQueries:
    def test_direct_service_property_query(self, service_type : ChemDBQueryExample, query_example : ChemDBQueryExample, expected_return : Any) -> None:
        '''Test if a chemical database query through a given service is executed completely and returns the expected result'''
        # prechecks to skip tests
        if not CHEMDB_STRATEGY_DEPENDENCIES_MET[service_type]:
            pytest.skip(f'{service_type.service_name} is missing Python dependencies')
            
        if not CHEMDB_STRATEGY_ONLINE[service_type]:
            pytest.skip(f'{service_type.service_name} cannot be connected to')
   
        # initialize and query service
        service = service_type()
        assert service.get_property(**asdict(query_example)) == expected_return
        
    def test_get_chemical_property_wrapper(self, service_type : ChemDBQueryExample, query_example : ChemDBQueryExample, expected_return : Any) -> None:
        '''Test that requests filtered through the get_chemical_properties() strategy wrapper are executed faithfully'''
        # prechecks to skip tests
        if not CHEMDB_STRATEGY_DEPENDENCIES_MET[service_type]:
            pytest.skip(f'{service_type.service_name} is missing Python dependencies')
            
        if not CHEMDB_STRATEGY_ONLINE[service_type]:
            pytest.skip(f'{service_type.service_name} cannot be connected to')
   
        # call get_chemical_property wrapper
        # try:
        print(asdict(query_example))
        assert get_chemical_property(**asdict(query_example), services=[service_type], fail_quietly=False) == expected_return # CRUCIAL that fail_quietly be False; rely on exceptions to match with xfails
        # except ChemicalDataQueryFailed:
            # print('foo')
    

# @pytest.mark.skipif(not CHEMDB_STRATEGIES_TO_TEST, reason=f'No chemical data server(s) is not available')
# def test_get_chemical_property() -> None:
#     a = 4


'''Utilities for generating, labelling, editing, and transforming RDKit molecules and other RDKit objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from .rdkdraw import (
    set_rdkdraw_size,
    enable_substruct_highlights,
    disable_substruct_highlights,
    enable_kekulized_drawing,
    disable_kekulized_drawing,
)